// simple temperture program + RF listening interface
/* Arduino pin out */
/*                  +--\/--+
      RESET     PC6 |      | PC5   ADC5/SCL   A5 (19)
 0    RXD       PD0 |  A   | PC4   ADC4/SDA   A4 (18)
 1    TXD       PD1 |  T   | PC3   ADC3       A3 (17)
 2    INT0      PD2 |  m   | PC2   ADC2       A2 (16)
 3 P  OC2B/INT1 PD3 |  e   | PC1   ADC1       A1 (15)
 4    XCK/T0    PD4 |  g   | PC0   ADC0       A0 (14)
                VCC |  a   | GND
                GND |  1   | AREF
      XTAL1     PB6 |  6   | AVCC
      XTAL2     PB7 |  8   | PB5   SCK        13
 5 P  OC0B/T1   PD5 |      | PB4   MISO       12
 6 P  OC0A/AIN0 PD6 |      | PB3   MOSI/OC2A  11 P
 7    AIN1      PD7 |      | PB2   SS/OC1B    10 P
 8    CLKO/ICP1 PB0 |      | PB1   PC1A        9 P
                    +------+
 
 --> termo input : 9
 --> 1-W I/O     : 8
 --> LED         : 14 (=A0, cannot use 13 since that is SPI SCK)
 */

#define RECEIVER
// #define SPI_DEBUG 1

// RF - we use interrupt-driven receiving but not sending
// That should be ok, because TX ints are used only to synchronize the write buffer which is sequential and thus easily pollable 

#if (defined __AVR_ATmega328P__) || (defined __AVR_ATmega168__) || (defined __AVR_ATmega168P__)
#define __ARDUINO 1
#elif (defined __AVR_ATmega644__) || (defined __AVR_ATmega644P__)
#define __SANGUINO 1
#else
#error Unsupported_MCU
#endif

#ifdef __ARDUINO
#define TERMO_PIN  9
#define W1_PIN     8 /* this definition is not really used - it is defined directly by port below */
#define LED_PIN   14

#define SCK_PIN   13
#define MISO_PIN  12
#define MOSI_PIN  11
#define SS_PIN    10
#define NIRQ_PIN   2 /* must be INT0 for the recevier mode */
#endif

#ifdef __SANGUINO
#define MOSI_PIN   5
#define MISO_PIN   6
#define SCK_PIN    7
#define SS_PIN     4
#define NIRQ_PIN   0 /* must be INT0 for the receiver mode */

#define TERMO_PIN 14
#define W1_PIN    15 /* this definition is not really used - it is defined directly by port below */
#define LED_PIN   13
#endif

const static char hv[] = "0123456789ABCDEF";

static void print_hex(const unsigned char *data, int len) {
   unsigned char i = 0;
   while (i < len) {
     Serial.print(hv[data[i] >> 4]);
     Serial.print(hv[data[i] & 15]);
     i++;
   }
}

//========================================= DS18S20 Temperature chip i/o

#if __SANGUINO
/* pin 15 on Sanguino = PD7  */
#define W1PORT PORTD
#define W1DDR  DDRD
#define W1PIN  PIND
#define W1BIT  PORTD7
#endif

#if __ARDUINO
/* pin 8 on Arduino */
#define W1PORT PORTB
#define W1DDR  DDRB
#define W1PIN  PINB
#define W1BIT  PORTB0
#endif

/* enable/disable the use of the internal pullup.
   It also uses output driver for the idle state - that may enhance stability but increases power usage.
   Note that the specs ask for 5k pullup whereas ATmega has 10k, but our timings are relaxed
   enough that it should still work in most situations.
   */
#define INT_PULLUP 1

#include <w1.h>

//========================================= RF

static void spi_setup() {
  volatile char tmp;
  pinMode(SCK_PIN, OUTPUT);
  pinMode(MOSI_PIN, OUTPUT);
  pinMode(MISO_PIN, INPUT);
  pinMode(SS_PIN, OUTPUT);
  pinMode(NIRQ_PIN, INPUT);
  digitalWrite(NIRQ_PIN, LOW);

  SPCR = 0;
  SPCR = _BV(SPR0) | _BV(SPE) | _BV(MSTR); // master, mode 0, MSB, FOSC/16 (1MHz)
  tmp = SPSR;
  tmp = SPDR;
}

static byte spi8(byte value)
{
  SPDR = value;
  while (!(SPSR & (_BV(SPIF)))) ;
  return SPDR;
}

static unsigned int spi16(unsigned int v) {
  unsigned int r;
  r = spi8(v >> 8) << 8;
  r |= spi8(v & 255);
  return r;
}

static volatile unsigned int rf_stat;

static unsigned int writeCmd(unsigned int v) {
  unsigned int r;
  digitalWrite(SS_PIN, LOW);
  r = spi16(v);
  if (!v) rf_stat = r;
  digitalWrite(SS_PIN, HIGH);
#ifdef SPI_DEBUG
  Serial.print("SPI cmd 0x"); Serial.print(hv[v >> 12]); Serial.print(hv[(v >> 8) & 15]); Serial.print(hv[(v >> 4) & 15]); Serial.print(hv[v & 15]);
  Serial.print(" -> 0x"); Serial.print(hv[r >> 12]); Serial.print(hv[(r >> 8) & 15]); Serial.print(hv[(r >> 4) & 15]); Serial.print(hv[r & 15]);
  Serial.print('\n');
#endif
  return r;
}

static char rf_can_send() { return (digitalRead(NIRQ_PIN) == LOW); }

static void rf_send(unsigned char v) {
  while (!rf_can_send()) ;
  writeCmd(0xB800 | v);
}

static void rf_send(const char *str) {
  while (*str) rf_send((unsigned char)*(str++));
}

static char rf_data_avail() { return ((writeCmd(0) & 0x8000) != 0); }

static unsigned char rf_recv() {
  unsigned int data;
  // while (!rf_data_avail());
  data = writeCmd(0xB000);
  return (data & 0xff);
}

/* modes for interrupt-driven operation */
#define RFM_NONE 0
#define RFM_SEND 1
#define RFM_RECV 2

static volatile byte rf_buf[64], rf_bufp;
static byte rf_mode;

static void rf_int() {
  unsigned int st = writeCmd(0);
  if (rf_mode == RFM_RECV) {
    if (st & 0x8000) { /* FIFO ready */
      if (rf_bufp < sizeof(rf_buf))
        rf_buf[rf_bufp++] = (byte) writeCmd(0xB000);
      else
        writeCmd(0xB000); /* we will simply discard data if the buffer overflows to avoid corruption */
    }
  }
}

void rf_fifo_rst() {
  writeCmd(0xCA81);
  writeCmd(0x0000);
  if (rf_mode == RFM_RECV)
    rf_bufp = 0;
  writeCmd(0xCA83);
}

#define RF_MODE_TRANSMITTER 0x20
#define RF_MODE_RECEIVER    0xc0

static void rf_init(unsigned char rf_on) {
  writeCmd(0x0000);
  writeCmd(0x80F7); //EL,EF,915MHz,12.0pF
  writeCmd(0x8219 | rf_on); //?er,!ebb,?et,es,ex,!eb,!ew,dc
  writeCmd(0xA7d0); //frequency select (7.5kHz steps, 2000=0x7d0 is 915MHz) 
  writeCmd(0xC647); //4.8kbps
  writeCmd(0x94A0); //VDI,FAST,134kHz,0dBm,-103dBm (0x94C0 for 67kHz)
  writeCmd(0xC2AC); //AL,!ml,DIG,DQD4
  writeCmd(0xCA81); //FIFO8,SYNC,!ff,DR
  writeCmd(0xCED4); //SYNC=2DD4A
  writeCmd(0xC443); //@PWR,NO RSTRIC,!st,!fi,OE,EN
  writeCmd(0x9850); //!mp,90kHz,MAX OUT
  writeCmd(0xCC77); //OB1ACOB0, LPX,Iddy,CDDIT,CBW0
  writeCmd(0xE000); //NOT USED
  writeCmd(0xC800); //NOT USED
  writeCmd(0xC040); //1.66MHz,2.2V
  if (rf_on == RF_MODE_RECEIVER) {
    rf_fifo_rst();
    attachInterrupt(0, rf_int, FALLING);
    rf_mode = RFM_RECV;
  }
}

//========================================= SETUP

void setup() {
  // serial
  Serial.begin(9600);
  Serial.println("minitemp 1.0-0");
  pinMode(LED_PIN, OUTPUT);
#ifdef RECEIVER
  pinMode(TERMO_PIN, INPUT);
  digitalWrite(TERMO_PIN, HIGH); /* enable pull-up */
  // 1-Wire
  w1_setup();
  // SPI
  spi_setup();
  rf_init(RF_MODE_RECEIVER);
#else
  spi_setup();
  rf_init(RF_MODE_TRANSMITTER);
#endif
}

//========================================= MAIN

static char termo = LOW;

static unsigned char address[16 * 8] = {
    0x28, 0x08, 0x2F, 0x18, 0x02, 0x00, 0x00, 0x43, /* #1 */
    0x28, 0xA1, 0x31, 0x18, 0x02, 0x00, 0x00, 0x57, /* #3  R  basement boiler */
    0x28, 0xD4, 0x2B, 0x18, 0x02, 0x00, 0x00, 0xE5, /* #2  Y  living room */
    0x28, 0x16, 0x23, 0x18, 0x02, 0x00, 0x00, 0x2A, /* #4  G  outside */
    0x28, 0xEF, 0x5C, 0x18, 0x02, 0x00, 0x00, 0xCC, /* #5  B  basement, desk */
    0x28, 0x0F, 0xDF, 0xF6, 0x01, 0x00, 0x00, 0xD8, /* #6  RY (built-in boiler) */
    0x28, 0x66, 0xB2, 0x8F, 0x02, 0x00, 0x00, 0xAF, /* #8  P  */
    0x28, 0xD1, 0x8B, 0x8F, 0x02, 0x00, 0x00, 0x5A, /* #9  RP */
    0x28, 0xBE, 0x8E, 0x8F, 0x02, 0x00, 0x00, 0x6F, /* #10 RG */
0 };
static unsigned char addresses = 8 * 9;

static unsigned char tsp[9];

static char rf_enabled = 1;

static void monitor_pin() {
  /* termo pin monitor */
  char cte = digitalRead(TERMO_PIN);
  if (cte != termo) {
    termo = cte;
    Serial.print("TS ");
    Serial.print(millis());
    Serial.println(cte ? " ON" : " OFF");
  }
}

#ifdef RECEIVER

static byte ibp=0; 

// this function is called while waiting
static void activeRun() {
    monitor_pin();
    
    while (ibp < rf_bufp) {
      byte r = rf_buf[ibp++];
      if (r == 0xAA || ibp == sizeof(rf_buf)) { /* end byte or end or buffer overflow */
        Serial.print("RF:");
        for (char i = 0; i < ibp; i++) {
          r = rf_buf[i]; 
          Serial.print(' '); Serial.print(hv[r >> 4]); Serial.print(hv[r & 15]);
        }
        Serial.print('\n');
        if (ibp == 17 && rf_buf[0] == 0x2d && rf_buf[1] == 0x87) { /* 8 bytes + 0xAA with leading 0x28 = T packet - try to decode */
          unsigned char d[8], crc = 0;
          for (char i = 0; i < 8; i++) crc = crc8(d[i] = (rf_buf[i << 1] & 0xf0) | ((rf_buf[(i << 1) | 1] & 0x0f) ^ 0x0f), crc);
          if (crc != 0) { Serial.println("--- T packet but invalid CRC"); }
          else {
            Serial.print('T'); for (char i = 0; i < 10; i++) Serial.print(hv[rf_buf[i] >> 4]); Serial.print(":\n");
            int temp = (int) ((((unsigned int) d[5]) << 8) | d[6]);
            if (temp < 0) { temp = -temp; Serial.print('-'); }
            Serial.print(temp / 100);
            Serial.print('.');
            temp %= 100;
            if (temp < 10) {
              Serial.print('0');
              Serial.println(((char)temp) + '0');
            } else {
              Serial.println(temp);
            }
          }
        } 
        rf_bufp = ibp = 0;
        rf_fifo_rst();
        break; /* explicit break even if the interrupt may have fed another byte */
      }
    }
}
#else
static void activeRun() {
  monitorPin();
}
#endif

static void activeDelay(unsigned int amount) {
  unsigned int st = millis(), et = st + amount;
  if (et > st) { /* easy part - no overflow */
    while (1) {
      activeRun();
      unsigned int ct = millis();
      if (ct > et || ct < st) return; /* the second condition is there just in case the et is on the edge of the overflow */
    }
  } else { /* overflow */
    while (1) {
      activeRun();
      unsigned int ct = millis();
      if (ct > et && ct < st) return; /* in the overflow case the time must fall between et and st */
    }
  }
  /* unreachable */
}

#ifndef RECEIVER
static void rf_do_send(const char *qq, char len) {
   Serial.println("RF:send");
   writeCmd(0x80F7);
   writeCmd(0x0000);
   writeCmd(0x8239);
   writeCmd(0x0000);
   rf_send(0xAA); // preamble
   rf_send(0xAA);
   rf_send(0xAA);
   rf_send(0x2D); // sync
   rf_send(0xD4);
   while (len--) { unsigned char o = *(qq++); rf_send((o & 0xf0) | (o >> 4) ^ 0x0F); rf_send((o << 4) | (o & 0x0f) ^ 0x0f); }
   rf_send(0xAA);
   rf_send(0xAA);
   rf_send(0xAA);
   writeCmd(0x8077); // disable TX
   writeCmd(0x8201); // disable everything
}
#endif


void loop() {
  digitalWrite(LED_PIN, HIGH);
  /* tell all devices to measure temp */
  if (!w1_talk_to(0, CONVERT_T)) { /* bother only if there is anything on the wire */
    /* the current code guarantees the bus to be in powered mode so parasite power is fine */
    activeDelay(800); /* wait 800ms for conversion */
    unsigned char q = 0;
    unsigned char send_buf[32];
    while (q < addresses) {
      w1_talk_to(address + q, READ_SP);
      tsp[0] = w1_read8();
      if (tsp[0] != 0xFF) { /* bother continuing only if it's a valid response */
        w1_read(tsp + 1, 8);
        Serial.print('T');
        print_hex(address + q, 8);
        Serial.println(": ");
        if (crc8(tsp, 9) == 0) { /* check CRC */
          if (tsp[0] == 0x50 && tsp[1] == 0x05) /* power-up default - we choose to ignore it for now */
            Serial.println("power-up default, ignoring");
          else {
            int temp = sp2temp(tsp);
#ifndef RECEIVER
            unsigned char crc = 0, *sh = send_buf;
            crc = crc8(*(sh++) = address[0], crc);
            crc = crc8(*(sh++) = address[1], crc);
            crc = crc8(*(sh++) = address[2], crc);
            crc = crc8(*(sh++) = address[3], crc);
            crc = crc8(*(sh++) = address[4], crc);
            crc = crc8(*(sh++) = (((unsigned int) temp) >> 8), crc);
            crc = crc8(*(sh++) = (((unsigned int) temp) & 255), crc);
            *(sh++) = crc;
            rf_do_send((const char*)send_buf, sh -send_buf);
 #endif
            if (temp < 0) { temp = -temp; Serial.print('-'); }
            Serial.print(temp / 100);
            Serial.print('.');
            temp %= 100;
            if (temp < 10) {
              Serial.print('0');
              Serial.println(((char)temp) + '0');
            } else {
              Serial.println(temp);
            }
          }
        } else Serial.println("invalid CRC");
      }
      q += 8;
    }
  } else Serial.println("(nothing on 1-wire)\n");
  w1_done(); /* de-power the bus */
  digitalWrite(LED_PIN, LOW);
 
  activeDelay(10000);
}
