// This is an extended version of the temperature query application
// Target: atmega644p
// pins: 13 (LED), 15,12,14 (DS18S20), 16-17 I2C (RTC), 24-30 (LCD)
// Interfaces RTC, shows time, date and temp on display and serial
// includes RTC tools. It can serve multiple DS pins.
//
//                   +---\/---+
//  INT0 (D 0) PB0  1|        |40  PA0 (AI 0 / D31)
//  INT1 (D 1) PB1  2|        |39  PA1 (AI 1 / D30)
//  INT2 (D 2) PB2  3|        |38  PA2 (AI 2 / D29)
//   PWM (D 3) PB3  4|        |37  PA3 (AI 3 / D28)
//   PWM (D 4) PB4  5|        |36  PA4 (AI 4 / D27)
//  MOSI (D 5) PB5  6|        |35  PA5 (AI 5 / D26)
//  MISO (D 6) PB6  7|        |34  PA6 (AI 6 / D25)
//   SCK (D 7) PB7  8|        |33  PA7 (AI 7 / D24)
//             RST  9|        |32  AREF
//             VCC 10|  644P  |31  GND 
//             GND 11|        |30  AVCC
//           XTAL2 12|        |29  PC7 (D 23)
//           XTAL1 13|        |28  PC6 (D 22)
//  RX0 (D 8)  PD0 14|        |27  PC5 (D 21) TDI
//  TX0 (D 9)  PD1 15|        |26  PC4 (D 20) TDO
//  RX1 (D 10) PD2 16|        |25  PC3 (D 19) TMS
//  TX1 (D 11) PD3 17|        |24  PC2 (D 18) TCKn
//  PWM (D 12) PD4 18|        |23  PC1 (D 17) SDA
//  PWM (D 13) PD5 19|        |22  PC0 (D 16) SCL
//  PWM (D 14) PD6 20|        |21  PD7 (D 15) PWM
//                   +--------+


// data pins:
// LCD: 24-30
// LED: 13 (status)
// DS18: 12, 14-15
// I2C: 16-17
// OUT - boiler: 23
// IN  - termostat: 22

#include <Wire.h>          // I2C
#include <OneWire.h>       // 1-Wire (DS18S20)
#include <LiquidCrystal.h> // LCD 4-bit

#define MAX_SCHED_TIME  5400 // do not allow schedules longer than 1.5 hours (even that is sort of extreme ...)

#define BOILER_PIN 23
#define TERMO_PIN  22

#define INT_PULLUP 1

LiquidCrystal lcd(24,25,26,27,28,29,30);

const byte my_addr   = 2;

const static char info_string[] = "MCT 1.0-3";

const static char hv[] = "0123456789ABCDEF";

volatile byte last_i2c_command = 0;
volatile int Tc_100 = 0; // last temp. measurement

static long time1;
static int  time2;

void i2c_receive(int nbytes) {
  Serial.print("I2C recv [");
  Serial.print(nbytes);
  Serial.print("]");
  for (int i = 0; i < nbytes; i++) {
    byte v = Wire.receive();
    if (i == 0) last_i2c_command = v;
    Serial.print(hv[v >> 4]);
    Serial.print(hv[v & 15]);
    Serial.print(' ');
  }
  Serial.print("\n");
}

void i2c_request(void) {
  Serial.print("I2C request!\n");
  return;
  if (last_i2c_command == 'T') {
    Wire.send((byte*)&Tc_100, (byte) 2);
    byte lo = Tc_100 & 255, hi = Tc_100 >> 8;
    Serial.print(">");
    Serial.print(hv[lo >> 4]); Serial.print(hv[lo & 15]);
    Serial.print(hv[hi >> 4]); Serial.print(hv[hi & 15]); Serial.print('\n');

  } else {
    int r = 0x55aa;
    Serial.print("?AA55\n");
    Wire.send((byte*)&r, (byte)2);
  }
}

// DS18S20 Temperature chip i/o

/* pin 15 on Sanguino = PD7  */
#define W1PORT PORTD
#define W1DDR  DDRD
#define W1PIN  PIND
#define W1BIT  PORTD7

/* enable/disable the use of the internal pullup.
   It also uses output driver for the idle state - that may enhance stability but increases power usage.
   Note that the specs ask for 5k pullup whereas ATmega has 10k, but our timings are relaxed
   enough that it should still work in most situations.
   */
#define INT_PULLUP 0

#define nop() __asm__("nop\n\t")
#ifndef cbi
#define cbi(sfr, bit) (_SFR_BYTE(sfr) &= ~_BV(bit))
#endif
#ifndef sbi
#define sbi(sfr, bit) (_SFR_BYTE(sfr) |= _BV(bit))
#endif 

void w1_setup() {
#if INT_PULLUP
  sbi(W1PORT, W1BIT); /* enable pullup */
#else
  cbi(W1PORT, W1BIT); /* disable pullup */
#endif
  cbi(W1DDR, W1BIT); /* init in idle (input) mode in either case - we'll be driving the bus on the reset pulse */
}

/* 0 on success */
static char w1_reset() {
  sbi(W1PORT, W1BIT); /* the bus should already be high, but just in case ... */
  sbi(W1DDR, W1BIT);
  /* let the device get a high pulse for charging */
  delayMicroseconds(10);
  /* pull 1W low for 480us */
  cbi(W1PORT, W1BIT);
  delayMicroseconds(480);
  sbi(W1PORT, W1BIT);
  delayMicroseconds(2); /* quick power-on in case the devices need it (this is not accordingh to specs but should help the recovery) */
  /* switch to input mode */
  cbi(W1DDR, W1BIT);
#if ! INT_PULLUP
  cbi(W1PORT, W1BIT); /* disable pull-up */
#endif
  /* wait until there is a presence pulse or at most 480us */
  unsigned long to = micros() + 480;
  char ps = 1;
  while ((ps = (W1PIN & _BV(W1BIT))) && micros() < to) ;
  if (!ps) { /* ok, pulse found */
    /* line is low,  wait for it to got back to high */
    to = micros() + 600; // extend the wait beyond 480 just in case some other device attempts to pull low ...
    while ((ps = (W1PIN & _BV(W1BIT))) == 0 && micros() < to) ;
    if (ps) {
      ps = 0; delayMicroseconds(15); /* just in case let the bus settle */
    } else ps = 1;
  }
  if (!ps) { /* power-up the bus if there are devices, otherwise we keep it unpowered */
    sbi(W1PORT, W1BIT); /* enable pull-up and put the bus back to driving mode */
    sbi(W1DDR, W1BIT);
  }
  return ps;
}

/* keeps the bus in powered state */
static void w1_write8(unsigned char val) {
  sbi(W1PORT, W1BIT); /* the bus should already be high, but just in case ... */
  sbi(W1DDR, W1BIT);
  for (char i = 0; i < 8; i++) {
    if (val & 1) {
      cbi(W1PORT, W1BIT); /* low */
      delayMicroseconds(4); /* this can be really anything from 4-8 */
      sbi(W1PORT, W1BIT); /* high */
      delayMicroseconds(65); /* the slot + recovery will be about 70us */  
    } else {
      cbi(W1PORT, W1BIT); /* low */
      delayMicroseconds(60); /* keep it low for the entire time of the slot */
      sbi(W1PORT, W1BIT); /* high */
      delayMicroseconds(10); /* recovery */
    }
    val >>= 1;
  }
}

static unsigned char w1_read8() {
  unsigned char val = 0;
  sbi(W1PORT, W1BIT);
  sbi(W1DDR, W1BIT); /* enable driver */
  for (char i = 0; i < 8; i++) {
      cbi(W1PORT, W1BIT); /* pull low */
      /* we need to wait 1us but delayMicroseconds gives us at best 4us .. we'll see */
      delayMicroseconds(2);
      cbi(W1DDR, W1BIT); /* back to input mode */
#if INT_PULLUP
      sbi(W1PORT, W1BIT); /* enable pull-up */
#endif
      /* the specs guarantee the value for 15us since we pulled the line low.
         For hard devices we could acually infer the case-specific timing from the ROM READ
         response, but it seems taking the edge case of 15us since *release* seems to work reasonably.
         Shorter times don't leave enough room for 1 bits to rise via pullup, longer times
         will sample after the slave has released the line. Fortunately both effects have usually
         the same rise time so the time could be adjusted according to the wiring/pullup characteristics.*/
      delayMicroseconds(17);
      val >>= 1;
      if (W1PIN & _BV(W1BIT)) val |= (unsigned char) 0x80;
      delayMicroseconds(43);
      sbi(W1PORT, W1BIT);
      sbi(W1DDR, W1BIT); /* enable output for recovery */
      delayMicroseconds(10);
  }
  return val;
}

/* this function puts the bus into idle state by de-powering the bus - i.e. releasing it to the pull-up */
static void w1_done() {
#if INT_PULLUP
  sbi(W1PORT, W1BIT);
#endif
  cbi(W1DDR, W1BIT); /* go to input more - releasing the bus */
#if ! INT_PULLUP
  cbi(W1PORT, W1BIT); /* remove internal pullup */
#endif
}

/* just a wrapper for read8 to read into a buffer */
static void w1_read(unsigned char *data, unsigned char len = 8)
{
  for (unsigned char i = 0; i < len; i++) data[i] = w1_read8();
}

/* CRC based on X^8 + X^5 + X^4 + 1 */
static unsigned char crc8(unsigned char data, unsigned char iv) {
  for (unsigned char i = 0; i < 8; i++) {
    iv = ((iv ^ data) & 1) ? ((iv >> 1) ^ 0x8C) : (iv >> 1);
    data >>= 1;
  }
  return iv;
}

/* same as above but will automatically feed an array with the initial value of 0 */
static unsigned char crc8(unsigned char *data, unsigned char len) {
  unsigned char iv = 0;
  for (unsigned char i = 0; i < len; i++) iv = crc8(data[i], iv);
  return iv;
}

/* converts 11-bit temperature from the DS18B20 tr into 1/100th of deg.C */
static int sp2temp(const unsigned char *data) {
  unsigned int iv = data[0] | (data[1] << 8), rv;
  char sign = iv & 0x8000;  // test most sig bit
  if (sign) iv = (~iv) + 1;
  rv = 6 * iv + iv / 4;
  return sign ? -rv : rv;
}

/* 1-Wire */
#define READ_ROM     0x33
#define SEARCH_ROM   0xF0
#define MATCH_ROM    0x55
#define SKIP_ROM     0xCC
/* DS18B20 */
#define ALARM_SEARCH 0xEC
#define CONVERT_T    0x44
#define WRITE_SP     0x4E /* TH, TL, CONF */
#define READ_SP      0xBE /* b0...b7 + CRC */

/* reset + match ROM + send command. if data is NULL then SKIP ROM is issued instead of MATCH ROM.
   if command is 0 then no command is sent.
   returns 1 is reset failed (no devices), 0 otherwise */
static int w1_talk_to(const unsigned char *addr, unsigned char cmd = 0) {
  if (w1_reset()) return 1;
  delayMicroseconds(60); // wait a bit after the presence pulse ... just in case ... 
  if (!addr)
    w1_write8(SKIP_ROM);
  else {
    w1_write8(MATCH_ROM);
    for (unsigned char i = 0; i < 8; i++) w1_write8(addr[i]);
  }
  if (cmd) w1_write8(cmd);
  return 0;
}

// NOTE: I2C read/write buffers are by default 32 bytes in Wire.h, so larger packets cannot be sent en-block

//-------------- RTC (DS1307) tools ----------------

#define RTC_ID ((byte) 0x68)

// checks RTC for hold bit and clears it if present. returns -1 on error, 0 if HC is clear, 1 if HC is set
// it should be used in setup() to make sure that the RTC is restarted ater a power failure
// it is also a good test whether the RTC chip is attached on I2C at all
int wakeRTC()
{
  int res = -1;
  byte hc = 0;
  Wire.beginTransmission(RTC_ID);
  Wire.send((byte) 0);
  if (Wire.endTransmission() != 0) return -1;
  if (Wire.requestFrom(RTC_ID, (byte) 1) != 1) return -1;
  if (Wire.available()) {
    hc = Wire.receive();
    res = (hc >> 7);      
  }
  if (res == 1) { // HC is set - resume the clock
    hc &= 0x3f;
    Wire.beginTransmission(RTC_ID);
    Wire.send((byte) 0);
    Wire.send(hc);
    Wire.endTransmission(); 
  }
  return res;
}

// read bytes from the RTC. The default length of 7 bytes (at offset 0) covers all RTC bytes
// note: uses a delay of 100ms
// returns # of bytes read on success, -1 on error
int readRTC(byte *rtc, int len = 7, int offset = 0) {
  int res = -1;
  Wire.beginTransmission(RTC_ID);
  Wire.send((byte) offset);
  if (Wire.endTransmission() != 0) return -1;
  res = (int) Wire.requestFrom(RTC_ID, (byte) len);
  if (res > 0 && Wire.available()) {
    unsigned int i = 0;
    while (i < len) rtc[i++] = Wire.receive();
  }
  return res;
}

// write bytes to the RTC. returns 0 on success, <0 on error (-1 len too big, -2/-3 NACK, -4 other error)
int writeRTC(byte *rtc, int len = 7, int offset = 0) {
  Wire.beginTransmission(RTC_ID);
  Wire.send((byte) offset);
  int i = 0;
  while (i < len) Wire.send(rtc[i]);
  return -((int) Wire.endTransmission());
}

#define HiD(X,padChar) (((X) > 15) ? ('0' + ((X) >> 4)) : (padChar))
#define LoD(X) (hv[(X) & 15])

#define RTC_DATE 1
#define RTC_TIME 2

// c must hold at least: 9 bytes for time, 11 bytes for date, 19 for both
void rtc2char(byte *rtc, char *c, int what = RTC_DATE | RTC_TIME, char pad_hour = ' ', char pad_date = 0)
{
  if (what & RTC_TIME) {
    if (rtc[2] > 15 || pad_hour) *(c++) = HiD(rtc[2], pad_hour);
    *(c++) = LoD(rtc[2]); *(c++) = ':';
    *(c++) = HiD(rtc[1], '0'); *(c++) = LoD(rtc[1]); *(c++) = ':';
    *(c++) = HiD(rtc[0], '0'); *(c++) = LoD(rtc[0]);
    if (what & RTC_DATE) *(c++) = ' ';
  }
  if (what & RTC_DATE) {
    if (rtc[4] > 15 || pad_date) *(c++) = HiD(rtc[4], pad_date);
    *(c++) = LoD(rtc[4]); *(c++) = '.';
    if (rtc[5] > 15 || pad_date) *(c++) = HiD(rtc[5], pad_date);
    *(c++) = LoD(rtc[5]); *(c++) = '.';    
    *(c++) = '2'; *(c++) = '0'; *(c++) = HiD(rtc[6], '0'); *(c++) = LoD(rtc[6]);
  }
  *c = 0;
}

/* pin 15 is now handled by non-search code */
#define DS_PINS 2
OneWire dss[DS_PINS] = { OneWire(12), OneWire(14) };

void printTime() {
  byte rtc[7];
  if (readRTC(rtc) > 0) {
    char tstr[24];
    rtc2char(rtc, tstr);
    Serial.print("RTC: "); Serial.println(tstr);  
    lcd.setCursor(0, 1); lcd.print(tstr);
  }
}

void setup() {
  // make sure the boiler pin is low on startup so it doesn't start the boiler accidentally
  pinMode(BOILER_PIN, OUTPUT);
  digitalWrite(BOILER_PIN, LOW);
  // also setup termostat input including 10k pull up
  pinMode(TERMO_PIN, INPUT);
  digitalWrite(TERMO_PIN, HIGH);
  // manual 1W setup
  w1_setup();
  // then setup serial
  Serial.begin(9600);
  Serial.println(info_string);
  // setup I2C
  Wire.begin(my_addr);
  Wire.onReceive(i2c_receive);
  Wire.onRequest(i2c_request);
  // setup LED
  pinMode(13, OUTPUT);
  Serial.println("I2C setup complete.\n");
  int rr = wakeRTC();
  Serial.println((rr == 1) ? "RTC with HC set, restarted" : ((rr == 0) ? "active RTC" : "no RTC"));
  lcd.print((rr == 1) ? "LCD enabled, HC reset" : "LCD enabled");
  printTime();
  Serial.print("t0: ");
  Serial.println(millis());
  lcd.setCursor(0, 2); 
}

#define MAXVALS 8

static unsigned int values[MAXVALS], value_base = 10, vwidth = 0, vv = 0;

static unsigned char addrb[8];
static unsigned char addrp = 0;

static void shiftValues() {
  for(unsigned int i = MAXVALS - 1; i > 0; i--) values[i] = values[i - 1];
  values[0] = 0; vwidth = 0; value_base = 10; addrp = 0; vv++;
}

static char termo = LOW;

static unsigned char address[16 * 8] = {
    0x28, 0xA1, 0x31, 0x18, 0x02, 0x00, 0x00, 0x57, /* #3  R  basement boiler */
    0x28, 0xD4, 0x2B, 0x18, 0x02, 0x00, 0x00, 0xE5, /* #2  Y  living room */
    0x28, 0x0F, 0xDF, 0xF6, 0x01, 0x00, 0x00, 0xD8, /* #6  RY (built-in boiler) */
    0x28, 0x66, 0xB2, 0x8F, 0x02, 0x00, 0x00, 0xAF, /* #8  P  */
    0x28, 0xD1, 0x8B, 0x8F, 0x02, 0x00, 0x00, 0x5A, /* #9  RP */
    0x28, 0xBE, 0x8E, 0x8F, 0x02, 0x00, 0x00, 0x6F, /* #10 RG */
0 };
static unsigned char addresses = 8 * 6;

static void update_time() {
   unsigned long ct = millis();
   if (ct < time1) /* overflow */
     time2++;
   time1 = ct;
}

static unsigned long s_time_adj = 0, b_time_on = 0, b_time_off = 0;

/* returns time in approximate seconds i.e. uses millis() / 1024 resolution instead of / 1000 for efficiency */
static unsigned long s2_time() {
  unsigned long c = time1;
  c >>= 10;
  c |= ((unsigned long) time2) << 22; 
  return c;
}

/* this function uses s_time_adj and corrects for the difference between 1024 internal base and 1000 real time.
   the adjustment requires extra 2 bits so it will fail if the system runs for more than 34.8 years --- unlikely so we should be safe. */
static unsigned long s_time() {
  unsigned long c = s2_time();
  unsigned long cadj = c; /* adj. is * ( 3 / 125 ) = ( 24 / 1000 ) ) */
  cadj *= 3;   /* x 3 */
  cadj /= 125; /* / 125 */
  return c + cadj + s_time_adj;
}

// this function is called while waiting
static void activeRun() {
  /* termo pin monitor */
  char cte = digitalRead(TERMO_PIN);
  if (cte != termo) {
    termo = cte;
    Serial.print("TS ");
    Serial.print(s_time());
    Serial.println(cte ? " ON" : " OFF");
    lcd.setCursor(0, 3);
    lcd.print(cte ? "T:ON " : "T:OFF");
    lcd.setCursor(0, 2);
  }
  update_time();
  /* boiler scheduling */
  unsigned long now = s2_time();
  if (b_time_on && b_time_on <= now) {
    b_time_on = 0;
    if (b_time_off > now) { /* we turn the boiler on ONLY if we have a valid off time */
      digitalWrite(BOILER_PIN, HIGH);
      Serial.print("B "); Serial.print(s_time()); Serial.println(" ON");
    } else {
      digitalWrite(BOILER_PIN, LOW); b_time_off = 0; /* just in case ... */
      Serial.println("ERROR: ON event without valid OFF, ignoring");
    }
  }
  if (b_time_off && b_time_off <= now) {
    digitalWrite(BOILER_PIN, LOW);
    Serial.print("B "); Serial.print(s_time()); Serial.println(" OFF");
    b_time_off = b_time_on = 0; /* off automatically resets on time as well */
  }
  
  /* serial commands processing */
  while (Serial.available() > 0) {
    int inc = Serial.read();
    if (inc == '#') { // '#' denotes the beginning of a 64-bit address
      value_base = 16; addrp = 1;
    } else if (inc == '$') { // "$" denotes hex mode and implies shift (unless width=0) to allow $aa$55
      if (vwidth != 0) shiftValues();
      value_base = 16;
    } else if (inc >= '0' && inc <= '9') { // 0-9 interpreted as decimal value
      values[0] *= value_base;
      values[0] += inc - 0x30;
      vwidth++;
      if (addrp && addrp < 9 && vwidth == 2) { addrb[(addrp++) - 1] = (unsigned char) values[0]; vwidth = 0; values[0] = 0; } 
    } else if (inc >= 0x61 && inc <= 0x66) { // a-f are interpreted as hex digits
      values[0] *= value_base;
      values[0] += inc - 0x61 + 10;
      vwidth++;
      if (addrp && addrp < 9 && vwidth == 2) { addrb[(addrp++) - 1] = (unsigned char) values[0]; vwidth = 0; values[0] = 0; } 
    } else if (inc == ',') { // shift values
      shiftValues();
    } else {
      if (vv > 0 || vwidth > 0) vv++;
      if (vv > MAXVALS) vv = MAXVALS; // overflow!
      switch (inc) {
        case 'T':
          if (addrp == 9) {
            unsigned char qp = 0;
            while (qp < addresses) {
             if (address[qp] == addrb[0] &&
                 address[qp + 1] == addrb[1] &&
                 address[qp + 2] == addrb[2] &&
                 address[qp + 3] == addrb[3] &&
                 address[qp + 4] == addrb[4] &&
                 address[qp + 5] == addrb[5] &&
                 address[qp + 6] == addrb[6] &&
                 address[qp + 7] == addrb[7]) break;
             qp += 8;
            }
            if (qp < addresses) Serial.println("OK already present.\n");
            else { /* new address */
              address[qp++] = addrb[0];               
              address[qp++] = addrb[1];
              address[qp++] = addrb[2];
              address[qp++] = addrb[3];
              address[qp++] = addrb[4];
              address[qp++] = addrb[5];
              address[qp++] = addrb[6];
              address[qp++] = addrb[7];
              addresses += 8;
            }
          } else Serial.println("ERR invalid address.\n");
          break;
        case 'H': // set pin to HIGH
        case 'L': // set pin to LOW
          if (vv > 0) {
            digitalWrite(values[--vv], (inc == 'H') ? HIGH : LOW);
            Serial.print("OK "); Serial.print(values[vv]); Serial.println((inc == 'H') ? " HIGH" : " LOW");
          } else Serial.println("ERROR: missing pin");
          break;
        case 'O':
        case 'I':
          if (vv > 0) {
            pinMode(values[--vv], (inc == 'O') ? OUTPUT : INPUT);
            Serial.print("OK "); Serial.print(values[vv]); Serial.println((inc == 'O') ? " OUT" : " IN");
          } else Serial.println("ERROR: missing pin");
          break;
        case 'A':
          if (vv > 0) {
            int a = analogRead(values[--vv]);
            Serial.print("OK "); Serial.print(values[vv]); Serial.print(' '); Serial.print(a); Serial.println(" A");
          } else Serial.println("ERROR: missing pin");
          break;
        case 'P':
          if (vv > 1) {
            int ap = values[--vv];
            int a = values[--vv];
            if (a < 0) a = 0; else if (a > 255) a = 255;
            analogWrite(ap, a);
            Serial.print("OK "); Serial.print(ap); Serial.print(' '); Serial.print(a); Serial.println(" PWD");
          } else Serial.println("ERROR: missing pin or value");
          break;
        case 'V':
          Serial.print("OK "); Serial.println(info_string);
          break;
        case 'W':
        {
          if (vv > 0) {
            Wire.beginTransmission((byte) values[--vv]);
            Serial.print("I2C WRITE ["); Serial.print(values[vv], HEX); Serial.print("]: ");
            while (vv) { Wire.send((byte) values[--vv]); Serial.print(values[vv], HEX); if (vv) Serial.print(','); }
            byte res = Wire.endTransmission();
            if (res != 0) {
              Serial.print(" ERROR: "); Serial.println((int) res);
            } else
              Serial.print(" OK\n");
          } else Serial.print("ERROR: no address specified\n");
        }
        break;
        case 'F':
          lcd.clear();
          break;
        case 'D':
          if (vv > 1) {
            unsigned int v1 = (unsigned int) values[--vv], v2 = (unsigned int) values[--vv];
            unsigned long t = v2; t <<= 16; t |= v1;
            if (t < 1265400000L || t > 1893400000 )
              Serial.println("ERROR: unplausible date/time");
            else {
              s_time_adj = 0; /* we need to subtract internal time -- to do that we have to calculate the current time so we do that without the adjustment */
              unsigned long ct = s_time();
              t -= ct;
              s_time_adj = t;
              Serial.print("OK "); Serial.print(ct); Serial.print(' '); Serial.println(t);
              lcd.setCursor(12, 3); lcd.print("POSIXt"); lcd.setCursor(0, 2);
            }
          } else Serial.println("ERROR: missing argument");
          break;
        case 'S': // schedule - times are from *now* in multiples of 1.024 seconds
          if (vv > 1) {
            b_time_on = values[--vv]; b_time_on += now;
            b_time_off = values[--vv]; b_time_off += now;
            if (b_time_off - b_time_on > MAX_SCHED_TIME) {
              b_time_on = b_time_off = 0;
              Serial.println("ERROR: invalid length to schedule");
            } else {
              Serial.print("OK "); Serial.print(now); Serial.print(' '); Serial.print(b_time_on); Serial.print(' '); Serial.println(b_time_off - b_time_on);
            }
          } else Serial.println("ERROR: missing argument");
          break;
        case 'R':
        {
          if (vv > 1) {
            byte addr = (byte) values[--vv];
            Serial.print("I2C READ ["); Serial.print(addr, HEX); Serial.print("] ");
            byte req = 1;
            if (vv > 0) req = values[--vv];
            if (req < 1) req = 1;
            Serial.print((int)req); Serial.print(" bytes, got ");
            int rec = (int) Wire.requestFrom(addr, req);
            rec &= 127;
            Serial.println(rec);
            while (rec > 0) {
              int inp = Wire.receive();
              Serial.print(inp, HEX);
              rec--;
              Serial.print((rec > 0) ? ',' : '\n');
            }
          } else Serial.print("ERROR: no address specified\n");
        }
        break;        
      }
      vv = 0; values[0] = 0; vwidth = 0; value_base = 10; addrp = 0;
    }
  }
}

static void activeDelay(unsigned long time) {
  unsigned long ct = millis();
  unsigned long tt = ct + time;
  if (tt < ct) { // overflow
    while (millis() > tt) activeRun();
  } else {
    while (millis() < tt) activeRun();
  }
}

static void print_hex(const unsigned char *data, int len) {
   unsigned char i = 0;
   while (i < len) {
     Serial.print(hv[data[i] >> 4]);
     Serial.print(hv[data[i] & 15]);
     i++;
   }
}

int cds = 0; // current DS
int err_count = 0;
#define MAX_ERR 3

static unsigned char tsp[9];

static void w1_doit() {
  /* tell all devices to measere temp */
  if (!w1_talk_to(0, CONVERT_T)) { /* bother only if there is anything on the wire */
    /* the current code guarantees the bus to be in powered mode so parasite power is fine */
    activeDelay(900); /* wait 900ms for conversion */
    unsigned char q = 0;
    while (q < addresses) {
      w1_talk_to(address + q, READ_SP);
      tsp[0] = w1_read8();
      if (tsp[0] != 0xFF) { /* bother continuing only if it's a valid response */
        w1_read(tsp + 1, 8);
        Serial.print('T');
        print_hex(address + q, 8);
        Serial.println(": ");
        if (crc8(tsp, 9) == 0) { /* check CRC */
          if (tsp[0] == 0x50 && tsp[1] == 0x05) /* power-up default - we choose to ignore it for now */
            Serial.println("power-up default, ignoring");
          else {  
            int temp = sp2temp(tsp);
            Serial.print(temp / 100);
            Serial.print('.');
            lcd.print(temp / 100);
            lcd.print('.');
            if (temp < 0) temp = -temp;
            temp %= 100;
            if (temp < 10) {
              Serial.print('0');  // lcd.print('0');
              Serial.println(((char)temp) + '0'); // lcd.print(((char)temp) + '0');
            } else {
              Serial.println(temp); // lcd.print(temp);
            }
            lcd.print(temp / 10); // use only 1 decimal point for LCD
            lcd.print(' ');
          }
        } else Serial.println("invalid CRC");
      }
      q += 8;
    }
  } else Serial.println("(nothing on pin 15)\n");
  w1_done(); /* de-power the bus */
}

void loop(void) {
  int HighByte, LowByte, TReading, SignBit, Whole, Fract;
  byte i;
  byte present = 0;
  byte data[12];
  byte addr[8];

  OneWire *ds = &dss[cds];
  if (err_count > MAX_ERR || !ds->search(addr)) { // current DS is done or out of control, try another one
      cds++;
      if (cds == DS_PINS) { // all DSes enumerated, take a break
        digitalWrite(13, HIGH);
        w1_doit();
        digitalWrite(13, LOW);
        cds = 0;
        activeDelay(10000); // don't repeat it too often or they will get warm
        printTime();
        lcd.setCursor(0, 2);
      }
      err_count = 0;
      dss[cds].reset_search();
      return;
  }

  Serial.print('T');
  for(i = 0; i < 8; i++) {
    if (addr[i] < 16) Serial.print("0");
    Serial.print(addr[i], HEX);
    //Serial.print(" ");
  }
  Serial.print(": \n");

  if ( OneWire::crc8( addr, 7) != addr[7]) {
      Serial.print("CRC is not valid!\n");
      err_count++;
      return;
  }

  if ( addr[0] != 0x28) {
      Serial.print("Device is not a DS18B20 family device.\n");
      err_count++;
      return;
  }

  digitalWrite(13, HIGH);
  ds->reset();
  ds->select(addr);
  ds->write(0x44,1);         // start conversion, with parasite power on at the end

  activeDelay(1000);     // maybe 750ms is enough, maybe not
  // we might do a ds.depower() here, but the reset will take care of it.

  present = ds->reset();
  ds->select(addr);    
  ds->write(0xBE);         // Read Scratchpad

  for ( i = 0; i < 9; i++)           // we need 9 bytes
    data[i] = ds->read();

  if (OneWire::crc8( data, 8) == data[8]) { /* proceed only if the CRC is correct */
  LowByte = data[0];
  HighByte = data[1];
  TReading = (HighByte << 8) + LowByte;
  SignBit = TReading & 0x8000;  // test most sig bit
  if (SignBit) // negative
  {
    TReading = (TReading ^ 0xffff) + 1; // 2's comp
  }
  Tc_100 = (6 * TReading) + TReading / 4;    // multiply by (100 * 0.0625) or 6.25

  Whole = Tc_100 / 100;  // separate off the whole and fractional portions
  Fract = Tc_100 % 100;

  if (SignBit) // If its negative
  {
     Tc_100 = -Tc_100; // make Tc negative for use in transmission
     Serial.print("-");
  }
  Serial.print(Whole);
  Serial.print(".");
  if (Fract < 10)
  {
     Serial.print("0");
  }
  Serial.print(Fract);

  Serial.print("\n");
  
  if (SignBit) lcd.print('-');
  lcd.print(Whole); lcd.print('.');
  //if (Fract < 10) lcd.print('0');
  lcd.print(Fract / 10);
  lcd.print(' ');
  } else
    Serial.println("CRC invalid\n");  
  digitalWrite(13, LOW);
}
